module Morphir.TestCasesJson exposing (..)

import Dict
import Expect exposing (Expectation)
import Json.Decode as Decode
import Morphir.Correctness.Codec exposing (decodeTestSuite, encodeTestSuite)
import Morphir.Correctness.Test exposing (TestCase, TestSuite)
import Morphir.IR as IR exposing (IR)
import Morphir.IR.Distribution.Codec as DistributionCodec
import Morphir.IR.Literal exposing (Literal(..))
import Morphir.IR.Value exposing (RawValue, Value(..))
import Test exposing (Test, test)


ir : IR
ir =
    Decode.decodeString DistributionCodec.decodeDistribution encodedDistribution
        |> Result.map IR.fromDistribution
        |> Result.withDefault IR.empty


expectedTestSuite : TestSuite
expectedTestSuite =
    Dict.fromList
        [ ( ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "limit", "tracking" ] )
          , [ { description = "List of Tracking Advantage"
              , expectedOutput = List () []
              , inputs =
                    [ Literal () (FloatLiteral 0)
                    , Literal () (FloatLiteral 5.5)
                    , Literal () (FloatLiteral 0)
                    , Literal () (FloatLiteral 12)
                    , Literal () (FloatLiteral 13)
                    ]
              }
            , { description = "List of Tracking Advantage"
              , expectedOutput =
                    List ()
                        [ Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "up" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "123456789") )
                            , ( [ "velocity" ], Literal () (FloatLiteral -13) )
                            ]
                        , Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "down" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "987654321") )
                            , ( [ "velocity" ], Literal () (FloatLiteral 6.6) )
                            ]
                        ]
              , inputs =
                    [ Literal () (FloatLiteral 10)
                    , Literal () (FloatLiteral 5.5)
                    , Literal () (FloatLiteral 10)
                    , Literal () (FloatLiteral 12)
                    , Literal () (FloatLiteral 13)
                    ]
              }
            , { description = "List of Tracking Advantage"
              , expectedOutput =
                    List ()
                        [ Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "up" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "123456789") )
                            , ( [ "velocity" ], Literal () (FloatLiteral 6.6) )
                            ]
                        , Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "down" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "987654321") )
                            , ( [ "velocity" ], Literal () (FloatLiteral 13) )
                            ]
                        , Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "up" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "123456789") )
                            , ( [ "velocity" ], Literal () (FloatLiteral 12) )
                            ]
                        , Record ()
                            [ ( [ "direction" ]
                              , Constructor () ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "insight" ], [ "use", "case", "1" ] ], [ "down" ] )
                              )
                            , ( [ "code" ], Literal () (StringLiteral "987654321") )
                            , ( [ "velocity" ], Literal () (FloatLiteral 12) )
                            ]
                        ]
              , inputs =
                    [ Literal () (FloatLiteral -10)
                    , Literal () (FloatLiteral -5.5)
                    , Literal () (FloatLiteral 10)
                    , Literal () (FloatLiteral 12)
                    , Literal () (FloatLiteral -13)
                    ]
              }
            ]
          )
        , ( ( [ [ "morphir" ], [ "reference" ], [ "model" ] ], [ [ "issues" ], [ "issue", "410" ] ], [ "add", "function" ] )
          , [ { description = "Add"
              , expectedOutput = Literal () (IntLiteral 9)
              , inputs = [ Literal () (IntLiteral 4), Literal () (IntLiteral 5) ]
              }
            , { description = "Add"
              , expectedOutput = Literal () (IntLiteral 29)
              , inputs = [ Literal () (IntLiteral 14), Literal () (IntLiteral 15) ]
              }
            , { description = "Add"
              , expectedOutput = Literal () (IntLiteral 38)
              , inputs = [ Literal () (IntLiteral 12), Literal () (IntLiteral 26) ]
              }
            , { description = "Add"
              , expectedOutput = Literal () (IntLiteral 248)
              , inputs = [ Literal () (IntLiteral 122), Literal () (IntLiteral 126) ]
              }
            , { description = "Add"
              , expectedOutput = Literal () (IntLiteral 11)
              , inputs = [ Literal () (IntLiteral 10), Literal () (IntLiteral 1) ]
              }
            , { description = "Add"
              , expectedOutput = Literal () (IntLiteral -9)
              , inputs = [ Literal () (IntLiteral -10), Literal () (IntLiteral 1) ]
              }
            ]
          )
        ]


evaluateJsonTest : Test
evaluateJsonTest =
    test "Encoded Test suite" <|
        \_ ->
            case encodeTestSuite ir expectedTestSuite of
                Ok encodedValue ->
                    case Decode.decodeValue (decodeTestSuite ir) encodedValue of
                        Ok testsuite ->
                            testsuite |> Expect.equal expectedTestSuite

                        Err error ->
                            Expect.fail (Debug.toString error)

                Err error ->
                    Expect.fail (Debug.toString error)

encodedDistribution : String
encodedDistribution =
     """
             [
                     "library",
                     [
                         [
                             "morphir"
                         ],
                         [
                             "reference"
                         ],
                         [
                             "model"
                         ]
                     ],
                     [],
                     {
                         "modules": [
                             {
                                 "name": [
                                     [
                                         "insight"
                                     ],
                                     [
                                         "use",
                                         "case",
                                         "1"
                                     ]
                                 ],
                                 "def": [
                                     "public",
                                     {
                                         "types": [
                                             [
                                                 [
                                                     "direction"
                                                 ],
                                                 [
                                                     "public",
                                                     [
                                                         "",
                                                         [
                                                             "custom_type_definition",
                                                             [],
                                                             [
                                                                 "public",
                                                                 [
                                                                     [
                                                                         [
                                                                             "down"
                                                                         ],
                                                                         []
                                                                     ],
                                                                     [
                                                                         [
                                                                             "up"
                                                                         ],
                                                                         []
                                                                     ]
                                                                 ]
                                                             ]
                                                         ]
                                                     ]
                                                 ]
                                             ],
                                             [
                                                 [
                                                     "tracking",
                                                     "advantage"
                                                 ],
                                                 [
                                                     "public",
                                                     [
                                                         "",
                                                         [
                                                             "type_alias_definition",
                                                             [],
                                                             [
                                                                 "record",
                                                                 {},
                                                                 [
                                                                     [
                                                                         [
                                                                             "direction"
                                                                         ],
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "reference"
                                                                                     ],
                                                                                     [
                                                                                         "model"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "insight"
                                                                                     ],
                                                                                     [
                                                                                         "use",
                                                                                         "case",
                                                                                         "1"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "direction"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "code"
                                                                         ],
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "string"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "string"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "velocity"
                                                                         ],
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "basics"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "float"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ]
                                                                     ]
                                                                 ]
                                                             ]
                                                         ]
                                                     ]
                                                 ]
                                             ]
                                         ],
                                         "values": [
                                             [
                                                 [
                                                     "limit",
                                                     "tracking"
                                                 ],
                                                 [
                                                     "public",
                                                     {
                                                         "inputTypes": [
                                                             [
                                                                 [
                                                                     "focal",
                                                                     "u",
                                                                     "r",
                                                                     "v"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "focal",
                                                                     "d",
                                                                     "r",
                                                                     "v"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "bet",
                                                                     "deviation",
                                                                     "threshold"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "point",
                                                                     "u",
                                                                     "r",
                                                                     "v"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "point",
                                                                     "d",
                                                                     "r",
                                                                     "v"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ]
                                                         ],
                                                         "outputType": [
                                                             "reference",
                                                             {},
                                                             [
                                                                 [
                                                                     [
                                                                         "morphir"
                                                                     ],
                                                                     [
                                                                         "s",
                                                                         "d",
                                                                         "k"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     [
                                                                         "list"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "list"
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "reference"
                                                                             ],
                                                                             [
                                                                                 "model"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "insight"
                                                                             ],
                                                                             [
                                                                                 "use",
                                                                                 "case",
                                                                                 "1"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "tracking",
                                                                             "advantage"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ]
                                                         ],
                                                         "body": [
                                                             "let_definition",
                                                             [
                                                                 "reference",
                                                                 {},
                                                                 [
                                                                     [
                                                                         [
                                                                             "morphir"
                                                                         ],
                                                                         [
                                                                             "s",
                                                                             "d",
                                                                             "k"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "list"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         "list"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "reference"
                                                                                 ],
                                                                                 [
                                                                                     "model"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "insight"
                                                                                 ],
                                                                                 [
                                                                                     "use",
                                                                                     "case",
                                                                                     "1"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "tracking",
                                                                                 "advantage"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ]
                                                                 ]
                                                             ],
                                                             [
                                                                 "bet",
                                                                 "rocket",
                                                                 "volume"
                                                             ],
                                                             {
                                                                 "inputTypes": [],
                                                                 "outputType": [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "float"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 "body": [
                                                                     "apply",
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "basics"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "float"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ],
                                                                     [
                                                                         "apply",
                                                                         [
                                                                             "function",
                                                                             {},
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "float"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ],
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "float"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "reference",
                                                                             [
                                                                                 "function",
                                                                                 {},
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "basics"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "float"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ],
                                                                                 [
                                                                                     "function",
                                                                                     {},
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "float"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "float"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "basics"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "add"
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "variable",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "float"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ],
                                                                             [
                                                                                 "focal",
                                                                                 "u",
                                                                                 "r",
                                                                                 "v"
                                                                             ]
                                                                         ]
                                                                     ],
                                                                     [
                                                                         "variable",
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "basics"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "float"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ],
                                                                         [
                                                                             "focal",
                                                                             "d",
                                                                             "r",
                                                                             "v"
                                                                         ]
                                                                     ]
                                                                 ]
                                                             },
                                                             [
                                                                 "let_definition",
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "list"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "list"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "reference"
                                                                                     ],
                                                                                     [
                                                                                         "model"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "insight"
                                                                                     ],
                                                                                     [
                                                                                         "use",
                                                                                         "case",
                                                                                         "1"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "tracking",
                                                                                     "advantage"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ]
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "interstellar",
                                                                     "tracking"
                                                                 ],
                                                                 {
                                                                     "inputTypes": [],
                                                                     "outputType": [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "list"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "list"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "reference"
                                                                                         ],
                                                                                         [
                                                                                             "model"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "insight"
                                                                                         ],
                                                                                         [
                                                                                             "use",
                                                                                             "case",
                                                                                             "1"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "tracking",
                                                                                         "advantage"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ]
                                                                         ]
                                                                     ],
                                                                     "body": [
                                                                         "if_then_else",
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "list"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "reference"
                                                                                             ],
                                                                                             [
                                                                                                 "model"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "insight"
                                                                                             ],
                                                                                             [
                                                                                                 "use",
                                                                                                 "case",
                                                                                                 "1"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "tracking",
                                                                                             "advantage"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "apply",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "bool"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ],
                                                                             [
                                                                                 "apply",
                                                                                 [
                                                                                     "function",
                                                                                     {},
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "bool"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "bool"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "reference",
                                                                                     [
                                                                                         "function",
                                                                                         {},
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "basics"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "or"
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "bool"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "equal"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "variable",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "focal",
                                                                                                 "u",
                                                                                                 "r",
                                                                                                 "v"
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "literal",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "float_literal",
                                                                                             0
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "apply",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "basics"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "bool"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "function",
                                                                                         {},
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "reference",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "equal"
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "variable",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "focal",
                                                                                             "d",
                                                                                             "r",
                                                                                             "v"
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "literal",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "float"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "float_literal",
                                                                                         0
                                                                                     ]
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "reference",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference"
                                                                                                 ],
                                                                                                 [
                                                                                                     "model"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "insight"
                                                                                                 ],
                                                                                                 [
                                                                                                     "use",
                                                                                                     "case",
                                                                                                     "1"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "tracking",
                                                                                                 "advantage"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "reference"
                                                                                     ],
                                                                                     [
                                                                                         "model"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "insight"
                                                                                     ],
                                                                                     [
                                                                                         "use",
                                                                                         "case",
                                                                                         "1"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "no",
                                                                                     "tracking",
                                                                                     "advantage"
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "if_then_else",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference"
                                                                                                 ],
                                                                                                 [
                                                                                                     "model"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "insight"
                                                                                                 ],
                                                                                                 [
                                                                                                     "use",
                                                                                                     "case",
                                                                                                     "1"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "tracking",
                                                                                                 "advantage"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "apply",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "basics"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "bool"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "function",
                                                                                         {},
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "reference",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "greater",
                                                                                                 "than"
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "variable",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "bet",
                                                                                             "rocket",
                                                                                             "volume"
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "literal",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "float"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "float_literal",
                                                                                         0
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "list",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "direction"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "model"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "insight"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "use",
                                                                                                                     "case",
                                                                                                                     "1"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "direction"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "up"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "literal",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "string"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "string_literal",
                                                                                                     "123456789"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "negate"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "variable",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "point",
                                                                                                     "d",
                                                                                                     "r",
                                                                                                     "v"
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "direction"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "model"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "insight"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "use",
                                                                                                                     "case",
                                                                                                                     "1"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "direction"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "down"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "literal",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "string"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "string_literal",
                                                                                                     "987654321"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "multiply"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "focal",
                                                                                                         "d",
                                                                                                         "r",
                                                                                                         "v"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "divide"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "variable",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "point",
                                                                                                             "u",
                                                                                                             "r",
                                                                                                             "v"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "focal",
                                                                                                         "u",
                                                                                                         "r",
                                                                                                         "v"
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "list",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "direction"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "model"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "insight"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "use",
                                                                                                                     "case",
                                                                                                                     "1"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "direction"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "up"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "literal",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "string"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "string_literal",
                                                                                                     "123456789"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "multiply"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "focal",
                                                                                                         "d",
                                                                                                         "r",
                                                                                                         "v"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "divide"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "variable",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "point",
                                                                                                             "u",
                                                                                                             "r",
                                                                                                             "v"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "focal",
                                                                                                         "u",
                                                                                                         "r",
                                                                                                         "v"
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "direction"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "constructor",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "model"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "insight"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "use",
                                                                                                                     "case",
                                                                                                                     "1"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "direction"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "down"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "literal",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "string"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "string"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "string_literal",
                                                                                                     "987654321"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "negate"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "variable",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "point",
                                                                                                     "d",
                                                                                                     "r",
                                                                                                     "v"
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ]
                                                                         ]
                                                                     ]
                                                                 },
                                                                 [
                                                                     "let_definition",
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "list"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "list"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "reference"
                                                                                         ],
                                                                                         [
                                                                                             "model"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "insight"
                                                                                         ],
                                                                                         [
                                                                                             "use",
                                                                                             "case",
                                                                                             "1"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "tracking",
                                                                                         "advantage"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ]
                                                                         ]
                                                                     ],
                                                                     [
                                                                         "collision",
                                                                         "tracking"
                                                                     ],
                                                                     {
                                                                         "inputTypes": [],
                                                                         "outputType": [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "list"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "reference"
                                                                                             ],
                                                                                             [
                                                                                                 "model"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "insight"
                                                                                             ],
                                                                                             [
                                                                                                 "use",
                                                                                                 "case",
                                                                                                 "1"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "tracking",
                                                                                             "advantage"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         "body": [
                                                                             "if_then_else",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference"
                                                                                                 ],
                                                                                                 [
                                                                                                     "model"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "insight"
                                                                                                 ],
                                                                                                 [
                                                                                                     "use",
                                                                                                     "case",
                                                                                                     "1"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "tracking",
                                                                                                 "advantage"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "apply",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "basics"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "bool"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "function",
                                                                                         {},
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "reference",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "or"
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "equal"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "variable",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "bet",
                                                                                                     "deviation",
                                                                                                     "threshold"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "literal",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "float_literal",
                                                                                                 0
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "bool"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "or"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "bool"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "bool"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "and"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "bool"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "bool"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "greater",
                                                                                                                     "than"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bet",
                                                                                                                 "deviation",
                                                                                                                 "threshold"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "float_literal",
                                                                                                             0
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "bool"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "greater",
                                                                                                                 "than"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "variable",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "bet",
                                                                                                             "rocket",
                                                                                                             "volume"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "literal",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "float_literal",
                                                                                                         0
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "and"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "bool"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "less",
                                                                                                                 "than"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "variable",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "bet",
                                                                                                             "deviation",
                                                                                                             "threshold"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "literal",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "float_literal",
                                                                                                         0
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "bool"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "less",
                                                                                                             "than"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "bet",
                                                                                                         "rocket",
                                                                                                         "volume"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "literal",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "float_literal",
                                                                                                     0
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "reference",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "reference"
                                                                                         ],
                                                                                         [
                                                                                             "model"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "insight"
                                                                                         ],
                                                                                         [
                                                                                             "use",
                                                                                             "case",
                                                                                             "1"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "no",
                                                                                         "tracking",
                                                                                         "advantage"
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "if_then_else",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "apply",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "basics"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "bool"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "function",
                                                                                             {},
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "bool"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "bool"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "greater",
                                                                                                     "than"
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "variable",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "bet",
                                                                                                 "rocket",
                                                                                                 "volume"
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "literal",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "float"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "float_literal",
                                                                                             0
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "if_then_else",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference"
                                                                                                         ],
                                                                                                         [
                                                                                                             "model"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "insight"
                                                                                                         ],
                                                                                                         [
                                                                                                             "use",
                                                                                                             "case",
                                                                                                             "1"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "tracking",
                                                                                                         "advantage"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "less",
                                                                                                         "than"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "abs"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "bet",
                                                                                                         "deviation",
                                                                                                         "threshold"
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "abs"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "variable",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "bet",
                                                                                                     "rocket",
                                                                                                     "volume"
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "up"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "123456789"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "down"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "987654321"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "up"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "123456789"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "down"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "987654321"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "if_then_else",
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference"
                                                                                                         ],
                                                                                                         [
                                                                                                             "model"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "insight"
                                                                                                         ],
                                                                                                         [
                                                                                                             "use",
                                                                                                             "case",
                                                                                                             "1"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "tracking",
                                                                                                         "advantage"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "apply",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "basics"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "bool"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "function",
                                                                                                 {},
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "bool"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bool"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "less",
                                                                                                         "than"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "abs"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "variable",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "bet",
                                                                                                         "deviation",
                                                                                                         "threshold"
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "apply",
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "float"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ],
                                                                                             [
                                                                                                 "reference",
                                                                                                 [
                                                                                                     "function",
                                                                                                     {},
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "s",
                                                                                                             "d",
                                                                                                             "k"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "basics"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "abs"
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "variable",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "s",
                                                                                                                 "d",
                                                                                                                 "k"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "basics"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "float"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "bet",
                                                                                                     "rocket",
                                                                                                     "volume"
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "up"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "123456789"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "down"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "987654321"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list",
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "up"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "123456789"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "apply",
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "function",
                                                                                                         {},
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "string"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "tracking",
                                                                                                                             "advantage"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "reference"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "model"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "insight"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "use",
                                                                                                                                 "case",
                                                                                                                                 "1"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "direction"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "string"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "string"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "reference"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "model"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "insight"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "use",
                                                                                                                                         "case",
                                                                                                                                         "1"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "tracking",
                                                                                                                                     "advantage"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "tracking",
                                                                                                                     "advantage"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "constructor",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "model"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "insight"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "use",
                                                                                                                             "case",
                                                                                                                             "1"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "direction"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "model"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "insight"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "use",
                                                                                                                         "case",
                                                                                                                         "1"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "down"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "literal",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "string"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "string"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "string_literal",
                                                                                                             "987654321"
                                                                                                         ]
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "apply",
                                                                                                     [
                                                                                                         "reference",
                                                                                                         {},
                                                                                                         [
                                                                                                             [
                                                                                                                 [
                                                                                                                     "morphir"
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "s",
                                                                                                                     "d",
                                                                                                                     "k"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     "basics"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "float"
                                                                                                             ]
                                                                                                         ],
                                                                                                         []
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "function",
                                                                                                             {},
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "multiply"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "abs"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "bet",
                                                                                                                     "deviation",
                                                                                                                     "threshold"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "apply",
                                                                                                         [
                                                                                                             "reference",
                                                                                                             {},
                                                                                                             [
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "morphir"
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "s",
                                                                                                                         "d",
                                                                                                                         "k"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         "basics"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "float"
                                                                                                                 ]
                                                                                                             ],
                                                                                                             []
                                                                                                         ],
                                                                                                         [
                                                                                                             "apply",
                                                                                                             [
                                                                                                                 "function",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 [
                                                                                                                     "function",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         "reference",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "morphir"
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "s",
                                                                                                                                     "d",
                                                                                                                                     "k"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     "basics"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "float"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         []
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "function",
                                                                                                                         {},
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "reference",
                                                                                                                             {},
                                                                                                                             [
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "morphir"
                                                                                                                                     ],
                                                                                                                                     [
                                                                                                                                         "s",
                                                                                                                                         "d",
                                                                                                                                         "k"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     [
                                                                                                                                         "basics"
                                                                                                                                     ]
                                                                                                                                 ],
                                                                                                                                 [
                                                                                                                                     "float"
                                                                                                                                 ]
                                                                                                                             ],
                                                                                                                             []
                                                                                                                         ]
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "divide"
                                                                                                                     ]
                                                                                                                 ]
                                                                                                             ],
                                                                                                             [
                                                                                                                 "variable",
                                                                                                                 [
                                                                                                                     "reference",
                                                                                                                     {},
                                                                                                                     [
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "morphir"
                                                                                                                             ],
                                                                                                                             [
                                                                                                                                 "s",
                                                                                                                                 "d",
                                                                                                                                 "k"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             [
                                                                                                                                 "basics"
                                                                                                                             ]
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "float"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     []
                                                                                                                 ],
                                                                                                                 [
                                                                                                                     "point",
                                                                                                                     "u",
                                                                                                                     "r",
                                                                                                                     "v"
                                                                                                                 ]
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "variable",
                                                                                                             [
                                                                                                                 "reference",
                                                                                                                 {},
                                                                                                                 [
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "morphir"
                                                                                                                         ],
                                                                                                                         [
                                                                                                                             "s",
                                                                                                                             "d",
                                                                                                                             "k"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         [
                                                                                                                             "basics"
                                                                                                                         ]
                                                                                                                     ],
                                                                                                                     [
                                                                                                                         "float"
                                                                                                                     ]
                                                                                                                 ],
                                                                                                                 []
                                                                                                             ],
                                                                                                             [
                                                                                                                 "focal",
                                                                                                                 "u",
                                                                                                                 "r",
                                                                                                                 "v"
                                                                                                             ]
                                                                                                         ]
                                                                                                     ]
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ]
                                                                         ]
                                                                     },
                                                                     [
                                                                         "apply",
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "list"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "reference"
                                                                                             ],
                                                                                             [
                                                                                                 "model"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "insight"
                                                                                             ],
                                                                                             [
                                                                                                 "use",
                                                                                                 "case",
                                                                                                 "1"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "tracking",
                                                                                             "advantage"
                                                                                         ]
                                                                                     ],
                                                                                     []
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "apply",
                                                                             [
                                                                                 "function",
                                                                                 {},
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "reference",
                                                                                 [
                                                                                     "function",
                                                                                     {},
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "s",
                                                                                                     "d",
                                                                                                     "k"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "reference",
                                                                                                 {},
                                                                                                 [
                                                                                                     [
                                                                                                         [
                                                                                                             "morphir"
                                                                                                         ],
                                                                                                         [
                                                                                                             "reference"
                                                                                                         ],
                                                                                                         [
                                                                                                             "model"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         [
                                                                                                             "insight"
                                                                                                         ],
                                                                                                         [
                                                                                                             "use",
                                                                                                             "case",
                                                                                                             "1"
                                                                                                         ]
                                                                                                     ],
                                                                                                     [
                                                                                                         "tracking",
                                                                                                         "advantage"
                                                                                                     ]
                                                                                                 ],
                                                                                                 []
                                                                                             ]
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "function",
                                                                                         {},
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "s",
                                                                                                         "d",
                                                                                                         "k"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "list"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "list"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "reference",
                                                                                                     {},
                                                                                                     [
                                                                                                         [
                                                                                                             [
                                                                                                                 "morphir"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "reference"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "model"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             [
                                                                                                                 "insight"
                                                                                                             ],
                                                                                                             [
                                                                                                                 "use",
                                                                                                                 "case",
                                                                                                                 "1"
                                                                                                             ]
                                                                                                         ],
                                                                                                         [
                                                                                                             "tracking",
                                                                                                             "advantage"
                                                                                                         ]
                                                                                                     ],
                                                                                                     []
                                                                                                 ]
                                                                                             ]
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "append"
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "variable",
                                                                                 [
                                                                                     "reference",
                                                                                     {},
                                                                                     [
                                                                                         [
                                                                                             [
                                                                                                 "morphir"
                                                                                             ],
                                                                                             [
                                                                                                 "s",
                                                                                                 "d",
                                                                                                 "k"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             [
                                                                                                 "list"
                                                                                             ]
                                                                                         ],
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "reference",
                                                                                             {},
                                                                                             [
                                                                                                 [
                                                                                                     [
                                                                                                         "morphir"
                                                                                                     ],
                                                                                                     [
                                                                                                         "reference"
                                                                                                     ],
                                                                                                     [
                                                                                                         "model"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     [
                                                                                                         "insight"
                                                                                                     ],
                                                                                                     [
                                                                                                         "use",
                                                                                                         "case",
                                                                                                         "1"
                                                                                                     ]
                                                                                                 ],
                                                                                                 [
                                                                                                     "tracking",
                                                                                                     "advantage"
                                                                                                 ]
                                                                                             ],
                                                                                             []
                                                                                         ]
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "interstellar",
                                                                                     "tracking"
                                                                                 ]
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "variable",
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "list"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "list"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "reference",
                                                                                         {},
                                                                                         [
                                                                                             [
                                                                                                 [
                                                                                                     "morphir"
                                                                                                 ],
                                                                                                 [
                                                                                                     "reference"
                                                                                                 ],
                                                                                                 [
                                                                                                     "model"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 [
                                                                                                     "insight"
                                                                                                 ],
                                                                                                 [
                                                                                                     "use",
                                                                                                     "case",
                                                                                                     "1"
                                                                                                 ]
                                                                                             ],
                                                                                             [
                                                                                                 "tracking",
                                                                                                 "advantage"
                                                                                             ]
                                                                                         ],
                                                                                         []
                                                                                     ]
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "collision",
                                                                                 "tracking"
                                                                             ]
                                                                         ]
                                                                     ]
                                                                 ]
                                                             ]
                                                         ]
                                                     }
                                                 ]
                                             ],
                                             [
                                                 [
                                                     "no",
                                                     "tracking",
                                                     "advantage"
                                                 ],
                                                 [
                                                     "public",
                                                     {
                                                         "inputTypes": [],
                                                         "outputType": [
                                                             "reference",
                                                             {},
                                                             [
                                                                 [
                                                                     [
                                                                         "morphir"
                                                                     ],
                                                                     [
                                                                         "s",
                                                                         "d",
                                                                         "k"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     [
                                                                         "list"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "list"
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "reference"
                                                                             ],
                                                                             [
                                                                                 "model"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "insight"
                                                                             ],
                                                                             [
                                                                                 "use",
                                                                                 "case",
                                                                                 "1"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "tracking",
                                                                             "advantage"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ]
                                                         ],
                                                         "body": [
                                                             "list",
                                                             [
                                                                 "reference",
                                                                 {},
                                                                 [
                                                                     [
                                                                         [
                                                                             "morphir"
                                                                         ],
                                                                         [
                                                                             "s",
                                                                             "d",
                                                                             "k"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "list"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         "list"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "reference"
                                                                                 ],
                                                                                 [
                                                                                     "model"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "insight"
                                                                                 ],
                                                                                 [
                                                                                     "use",
                                                                                     "case",
                                                                                     "1"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "tracking",
                                                                                 "advantage"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ]
                                                                 ]
                                                             ],
                                                             []
                                                         ]
                                                     }
                                                 ]
                                             ]
                                         ]
                                     }
                                 ]
                             },
                             {
                                 "name": [
                                     [
                                         "issues"
                                     ],
                                     [
                                         "issue",
                                         "410"
                                     ]
                                 ],
                                 "def": [
                                     "public",
                                     {
                                         "types": [],
                                         "values": [
                                             [
                                                 [
                                                     "add",
                                                     "function"
                                                 ],
                                                 [
                                                     "public",
                                                     {
                                                         "inputTypes": [
                                                             [
                                                                 [
                                                                     "num",
                                                                     "1"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "int"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "int"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ],
                                                             [
                                                                 [
                                                                     "num",
                                                                     "2"
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "int"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "int"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ]
                                                             ]
                                                         ],
                                                         "outputType": [
                                                             "reference",
                                                             {},
                                                             [
                                                                 [
                                                                     [
                                                                         "morphir"
                                                                     ],
                                                                     [
                                                                         "s",
                                                                         "d",
                                                                         "k"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     [
                                                                         "basics"
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "int"
                                                                 ]
                                                             ],
                                                             []
                                                         ],
                                                         "body": [
                                                             "apply",
                                                             [
                                                                 "reference",
                                                                 {},
                                                                 [
                                                                     [
                                                                         [
                                                                             "morphir"
                                                                         ],
                                                                         [
                                                                             "s",
                                                                             "d",
                                                                             "k"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             "basics"
                                                                         ]
                                                                     ],
                                                                     [
                                                                         "int"
                                                                     ]
                                                                 ],
                                                                 []
                                                             ],
                                                             [
                                                                 "apply",
                                                                 [
                                                                     "function",
                                                                     {},
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "basics"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "int"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ],
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "basics"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "int"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "reference",
                                                                     [
                                                                         "function",
                                                                         {},
                                                                         [
                                                                             "reference",
                                                                             {},
                                                                             [
                                                                                 [
                                                                                     [
                                                                                         "morphir"
                                                                                     ],
                                                                                     [
                                                                                         "s",
                                                                                         "d",
                                                                                         "k"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     [
                                                                                         "basics"
                                                                                     ]
                                                                                 ],
                                                                                 [
                                                                                     "int"
                                                                                 ]
                                                                             ],
                                                                             []
                                                                         ],
                                                                         [
                                                                             "function",
                                                                             {},
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "int"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ],
                                                                             [
                                                                                 "reference",
                                                                                 {},
                                                                                 [
                                                                                     [
                                                                                         [
                                                                                             "morphir"
                                                                                         ],
                                                                                         [
                                                                                             "s",
                                                                                             "d",
                                                                                             "k"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         [
                                                                                             "basics"
                                                                                         ]
                                                                                     ],
                                                                                     [
                                                                                         "int"
                                                                                     ]
                                                                                 ],
                                                                                 []
                                                                             ]
                                                                         ]
                                                                     ],
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "add"
                                                                         ]
                                                                     ]
                                                                 ],
                                                                 [
                                                                     "variable",
                                                                     [
                                                                         "reference",
                                                                         {},
                                                                         [
                                                                             [
                                                                                 [
                                                                                     "morphir"
                                                                                 ],
                                                                                 [
                                                                                     "s",
                                                                                     "d",
                                                                                     "k"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 [
                                                                                     "basics"
                                                                                 ]
                                                                             ],
                                                                             [
                                                                                 "int"
                                                                             ]
                                                                         ],
                                                                         []
                                                                     ],
                                                                     [
                                                                         "num",
                                                                         "1"
                                                                     ]
                                                                 ]
                                                             ],
                                                             [
                                                                 "variable",
                                                                 [
                                                                     "reference",
                                                                     {},
                                                                     [
                                                                         [
                                                                             [
                                                                                 "morphir"
                                                                             ],
                                                                             [
                                                                                 "s",
                                                                                 "d",
                                                                                 "k"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             [
                                                                                 "basics"
                                                                             ]
                                                                         ],
                                                                         [
                                                                             "int"
                                                                         ]
                                                                     ],
                                                                     []
                                                                 ],
                                                                 [
                                                                     "num",
                                                                     "2"
                                                                 ]
                                                             ]
                                                         ]
                                                     }
                                                 ]
                                             ]
                                         ]
                                     }
                                 ]
                             }
                         ]
                     }
                 ]
         """